<?php

declare(strict_types=1);

namespace Drupal\migrate_from_services\Form;

use Drupal\Core\Form\FormBase;
use Drupal\Core\Form\FormStateInterface;
use Drupal\migrate\MigrateMessage;
use Drupal\migrate\Plugin\MigrationPluginManagerInterface;
use Drupal\migrate_from_services\MigrateBatchExecutable;
use Drupal\migrate_tools\MigrateTools;
use GuzzleHttp\Client;
use Psr\Log\LoggerInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Build form to configure migrate execution pipeline.
 */
class MigrationExecuteForm extends FormBase {

  /**
   * Plugin manager for migration plugins.
   */
  protected MigrationPluginManagerInterface $migrationPluginManager;

  /**
   * The logger service.
   *
   * @var \Psr\Log\LoggerInterface
   */
  protected LoggerInterface $logger;

  /**
   * The http client.
   *
   * @var \GuzzleHttp\Client
   */
  protected Client $httpClient;

  /**
   * Constructs a new MigrationExecuteForm object.
   *
   * @param \Drupal\migrate\Plugin\MigrationPluginManagerInterface $migration_plugin_manager
   *   The plugin manager for config entity-based migrations.
   * @param \Psr\Log\LoggerInterface $logger
   *   The logger service.
   * @param \GuzzleHttp\ClientInterface $http_client
   *   The Guzzle HTTP client.
   */
  public function __construct(
    MigrationPluginManagerInterface $migration_plugin_manager,
    LoggerInterface $logger,
    Client $http_client,
  ) {
    $this->migrationPluginManager = $migration_plugin_manager;
    $this->logger = $logger;
    $this->httpClient = $http_client;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container): self {
    return new static(
      $container->get('plugin.manager.migration'),
      $container->get('logger.factory')->get('migrate_from_services'),
      $container->get('http_client')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function getFormId(): string {
    return 'migration_services_execute_form';
  }

  /**
   * {@inheritdoc}
   */
  public function buildForm(array $form, FormStateInterface $form_state): array {
    $form['data_source'] = [
      '#type' => 'select',
      '#title' => $this->t('Data source'),
      '#options' => [
        'url' => $this->t('Service URL'),
        'json' => $this->t('Raw JSON Text'),
        'manual' => $this->t('Manual'),
      ],
    ];

    $form['json_url'] = [
      '#type' => 'url',
      '#title' => $this->t('JSON service URL'),
      '#description' => $this->t("The migration will be executed by reading a JSON."),
      '#states' => [
        'visible' => [
          'select[name="data_source"]' => ['value' => 'url'],
        ],
      ],
    ];

    $form['json_text'] = [
      '#type' => 'textarea',
      '#title' => $this->t('Raw JSON Text'),
      '#description' => $this->t("The migration will be performed from a JSON with a valid format.
      This is useful for testing with different JSON."),
      '#size' => 1024,
      '#states' => [
        'visible' => [
          'select[name="data_source"]' => ['value' => 'json'],
        ],
      ],
    ];

    $form['migration_id'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Migration ID'),
      '#description' => $this->t('Migration ID defined in the migrate plugin (YAML file)'),
      '#states' => [
        'visible' => [
          'select[name="data_source"]' => ['value' => 'manual'],
        ],
      ],
    ];

    $form['idlist'] = [
      '#type' => 'textfield',
      '#title' => $this->t('ID List'),
      '#description' => $this->t('Comma-separated list of IDs to process.'),
      '#maxlength' => 255,
      '#size' => 60,
      '#pattern' => '^[0-9]+(' . MigrateTools::DEFAULT_ID_LIST_DELIMITER . '[0-9]+)?(,?[0-9]+(' . MigrateTools::DEFAULT_ID_LIST_DELIMITER . '[0-9]+)?)*$',
      '#states' => [
        'visible' => [
          'select[name="data_source"]' => ['value' => 'manual'],
        ],
      ],
    ];

    $form['limit'] = [
      '#type' => 'number',
      '#title' => $this->t('Limit'),
      '#description' => $this->t('Set a limit of how many items to process for each migration task.'),
      '#states' => [
        'visible' => [
          'select[name="data_source"]' => ['value' => 'manual'],
        ],
      ],
    ];

    $form['update'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Update existing records'),
      '#description' => $this->t('Check this box to update all previously-imported content in addition to importing new content. Leave unchecked to only import new content'),
      '#states' => [
        'visible' => [
          'select[name="data_source"]' => ['value' => 'manual'],
        ],
      ],
    ];

    $form['force'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Ignore dependencies'),
      '#description' => $this->t('Check this box to ignore dependencies when running imports - all tasks will run whether or not their dependent tasks have completed.'),
      '#states' => [
        'visible' => [
          'select[name="data_source"]' => ['value' => 'manual'],
        ],
      ],
    ];

    $json_sample_value = '
<pre>
{
  "migration_id": "migrate_node_article",
  "options": {
    "idlist": [12, 23, 45, 67, 89],
    "limit": 50,
    "update": "TRUE",
    "force": "FALSE"
  }
}
</pre>';
    $form['json_sample'] = [
      '#type' => 'details',
      '#title' => $this->t('Valid JSON format'),
      '#markup' => $json_sample_value,
      '#states' => [
        'invisible' => [
          'select[name="data_source"]' => ['value' => 'manual'],
        ],
      ],
    ];

    $form['submit'] = [
      '#type' => 'submit',
      '#value' => $this->t('Execute migration'),
    ];

    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function validateForm(array &$form, FormStateInterface $form_state): void {
    if ($form_state->getValue('data_source') == 'manual') {
      if (empty($form_state->getValue('migration_id'))) {
        $form_state->setErrorByName('migration_id', $this->t('Missing Migration ID value'));
      }
    }
  }

  /**
   * Fetch the JSON data.
   *
   * @param string $url
   *   The Service URL.
   *
   * @return mixed|null
   *   Return the JSON response decoded.
   *
   * @throws \GuzzleHttp\Exception\GuzzleException
   */
  private function fetchJsonData(string $url) {
    try {
      $response = $this->httpClient->get($url, ['timeout' => 10]);

      if ($response->getStatusCode() == 200) {
        $body = $response->getBody()->getContents();
        return json_decode($body, TRUE);
      }
    }
    catch (\Exception $e) {
      $this->logger->error($this->t('Error in the request: @message', ['@message' => $e->getMessage()]));
    }

    return NULL;
  }

  /**
   * {@inheritdoc}
   */
  public function submitForm(array &$form, FormStateInterface $form_state): void {
    $migration_id = '';
    $options = [];

    // Prepare migration data depends of data_source mode.
    if ($form_state->getValue('data_source') == 'url') {
      // JSON Service URL migration.
      $data = $this->fetchJsonData($form_state->getValue('json_url'));

      if ($data === NULL) {
        $this->logger->error($this->t('The JSON could not be obtained from the service provided.'));
        \Drupal::messenger()->addError($this->t('The JSON could not be obtained from the service provided.'));
        return;
      }

      // Validate JSON migration_id key.
      if (!isset($data['migration_id'])) {
        $this->logger->error($this->t('The JSON must contain the key “migration_id” with a valid value.'));
        \Drupal::messenger()->addError($this->t('The JSON must contain the key “migration_id” with a valid value.'));
        return;
      }

      $migration_id = $data['migration_id'];
      $options = [
        'idlist' => $data['options']['idlist'] ?? '',
        'limit' => $data['options']['limit'] ?? 0,
        'update' => $data['options']['update'] ?? 0,
        'force' => $data['options']['force'] ?? 0,
      ];
    }
    elseif ($form_state->getValue('data_source') == 'json') {
      // Raw JSON Text migration.
      $json = $form_state->getValue('json_text');
      $json = trim($json);
      $data = json_decode($json, TRUE);

      $migration_id = $data['migration_id'];
      $options = [
        'idlist' => $data['options']['idlist'] ?? '',
        'limit' => $data['options']['limit'] ?? 0,
        'update' => $data['options']['update'] ?? 0,
        'force' => $data['options']['force'] ?? 0,
      ];
    }
    else {
      // Manual migration.
      $migration_id = $form_state->getValue('migration_id');
      $options = $this->buildOptions($form_state);
    }

    $this->startMigrationBatch($migration_id, $options);
  }

  /**
   * Starts Batch import migration process.
   *
   * @param string $migration_id
   *   The migration id of plugin migrate.
   * @param array $options
   *   The options of migration pipeline.
   *
   * @return void
   *   Starts the batch import.
   *
   * @throws \Drupal\Component\Plugin\Exception\PluginException
   */
  private function startMigrationBatch(string $migration_id, array $options): void {
    /** @var \Drupal\migrate\Plugin\MigrationInterface $migration_plugin */
    $migration_plugin = $this->migrationPluginManager->createInstance($migration_id, ['migration' => $migration_id]);
    $migrateMessage = new MigrateMessage();

    $executable = new MigrateBatchExecutable($migration_plugin, $migrateMessage, $options);
    $executable->batchImport();
  }

  /**
   * Build migrate execute options from the submitted form values.
   *
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The form state.
   *
   * @return array
   *   Options array for the migrate execution.
   */
  protected function buildOptions(FormStateInterface $form_state) {
    $options = [
      'limit' => $form_state->getValue('limit') ?: 0,
      'update' => $form_state->getValue('update') ?: 0,
      'force' => $form_state->getValue('force') ?: 0,
    ];

    if ($idlist = $form_state->getValue('idlist')) {
      $options['idlist'] = $idlist;
    }

    return $options;
  }

}
